/*!
 * Copyright (C) 2006 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 */

/*!
 * $Log$
 */

#include <cfg/os.h>
#include <cfg/clock.h>
#include <dev/board.h>
#include <dev/irqreg.h>

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <sys/atom.h>
#include <sys/event.h>
#include <sys/timer.h>

#include "at73dac.h"

#ifndef DACSPI_ID
#if defined(AT91SAM7X_EK)       /* Board */
#define DACSPI_ID   SPI0_ID
#define DACSPI_PCS  SPI_PCS_2
#elif defined(AT91SAM9260_EK)   /* Board */
#define DACSPI_ID   SPI1_ID
#define DACSPI_PCS  SPI_PCS_0
#define DAC_LED_BIT 6
#else                           /* Board */
#warning "Unknown board configuration"
#endif                          /* Board */
#endif

#if defined(MCU_AT91SAM7X256)   /* MCU_ */

#define DACSPI_PIO_ID    PIOA_ID
#if DACSPI_ID == SPI0_ID
#define DACSPI_PINS_A  _BV(PA16_SPI0_MISO_A) | _BV(PA17_SPI0_MOSI_A) | _BV(PA18_SPI0_SPCK_A) | _BV(PA14_SPI0_NPCS2_A)
#define DACSPI_PINS_B  0
#elif DACSPI_ID == SPI1_ID
#define DACSPI_PINS_A  _BV(PA24_SPI1_MISO_B) | _BV(PA23_SPI1_MOSI_B) | _BV(PA22_SPI1_SPCK_B) | _BV(PA25_SPI1_NPCS1_B)
#define DACSPI_PINS_B  0
#endif

#define DACI2S_PIO_ID   PIOA_ID
#define DACI2S_PINS_A   _BV(PA23_TD_A) | _BV(PA21_TF_A) | _BV(PA22_TK_A)
#define DACI2S_PINS_B   0

#elif defined (MCU_AT91SAM9260) /* MCU_ */

#if DACSPI_ID == SPI0_ID
#define DACSPI_PIO_ID    PIOA_ID
#define DACSPI_PINS_A   _BV(PA0_SPI0_MISO_A) | _BV(PA1_SPI0_MOSI_A) | _BV(PA2_SPI0_SPCK_A) | _BV(PA3_SPI0_NPCS0_A)
#define DACSPI_PINS_B   0
#elif DACSPI_ID == SPI1_ID
#define DACSPI_PIO_ID    PIOB_ID
#define DACSPI_PINS_A   _BV(PB0_SPI1_MISO_A) | _BV(PB1_SPI1_MOSI_A) | _BV(PB2_SPI1_SPCK_A) | _BV(PB3_SPI1_NPCS0_A)
#define DACSPI_PINS_B   0
#endif

#define DACI2S_PIO_ID   PIOB_ID
#define DACI2S_PINS_A   _BV(PB18_TD0_A) | _BV(PB17_TF0_A) | _BV(PB16_TK0_A)
#define DACI2S_PINS_B   0

#else                           /* MCU_ */
#warning "Unknown MCU"
#endif                          /* MCU_ */

#if DACSPI_ID == SPI0_ID        /* DACSPI_ID */
#define DACSPI_MR   SPI0_MR
#define DACSPI_CR   SPI0_CR
#define DACSPI_SR   SPI0_SR
#define DACSPI_TDR  SPI0_TDR
#define DACSPI_RDR  SPI0_RDR
#if DACSPI_PCS == SPI_PCS_0
#define DACSPI_CSR  SPI0_CSR0
#elif DACSPI_PCS == SPI_PCS_1
#define DACSPI_CSR  SPI0_CSR1
#elif DACSPI_PCS == SPI_PCS_2
#define DACSPI_CSR  SPI0_CSR2
#elif DACSPI_PCS == SPI_PCS_3
#define DACSPI_CSR  SPI0_CSR3
#endif
#elif DACSPI_ID == SPI1_ID      /* DACSPI_ID */
#define DACSPI_MR   SPI1_MR
#define DACSPI_CR   SPI1_CR
#define DACSPI_SR   SPI1_SR
#define DACSPI_TDR  SPI1_TDR
#define DACSPI_RDR  SPI1_RDR
#if DACSPI_PCS == SPI_PCS_0
#define DACSPI_CSR  SPI1_CSR0
#elif DACSPI_PCS == SPI_PCS_1
#define DACSPI_CSR  SPI1_CSR1
#elif DACSPI_PCS == SPI_PCS_2
#define DACSPI_CSR  SPI1_CSR2
#elif DACSPI_PCS == SPI_PCS_3
#define DACSPI_CSR  SPI1_CSR3
#endif
#endif                          /* DACSPI_ID */

#if DACSPI_PIO_ID == PIOA_ID    /* DACSPI_PIO_ID */
#define DACSPI_PDR  PIOA_PDR
#define DACSPI_ASR  PIOA_ASR
#define DACSPI_BSR  PIOA_BSR
#elif DACSPI_PIO_ID == PIOB_ID  /* DACSPI_PIO_ID */
#define DACSPI_PDR  PIOB_PDR
#define DACSPI_ASR  PIOB_ASR
#define DACSPI_BSR  PIOB_BSR
#endif                          /* DACSPI_PIO_ID */

#if DACI2S_PIO_ID == PIOA_ID    /* DACI2S_PIO_ID */
#define DACI2S_PDR  PIOA_PDR
#define DACI2S_ASR  PIOA_ASR
#define DACI2S_BSR  PIOA_BSR
#elif DACI2S_PIO_ID == PIOB_ID  /* DACI2S_PIO_ID */
#define DACI2S_PDR  PIOB_PDR
#define DACI2S_ASR  PIOB_ASR
#define DACI2S_BSR  PIOB_BSR
#endif                          /* DACI2S_PIO_ID */


static HANDLE i2s_que;

#ifndef SAMPLE_BUFFERS
#if defined (AT91SAM9260_EK)
//#define SAMPLE_BUFFERS  256
#define SAMPLE_BUFFERS  32
#else
#define SAMPLE_BUFFERS  3
#endif
#endif

#define PCM_CHANS   2
#define PCM_BITS    16

typedef struct _WAVEBUFF {
    u_short *wbf_dat;
    int wbf_siz;
    int wbf_len;
} WAVEBUFF;

static WAVEBUFF wavbuf[SAMPLE_BUFFERS];
static volatile u_int brd_idx;
static u_int bwr_idx;

static u_int ovrsel;
static int spi_initialized;

/*!
 * \brief Set PDC pointer and counter registers.
 *
 * SSC interrupts must have been disabled befor calling this function.
 */
#if 0
static INLINE void I2sPdcFill(void)
{
    if (brd_idx != bwr_idx) {
        if (inr(SSC_TCR) == 0) {
            if (++brd_idx >= SAMPLE_BUFFERS) {
                brd_idx = 0;
            }
            outr(SSC_TPR, (u_int) wavbuf[brd_idx].wbf_dat);
            outr(SSC_TCR, wavbuf[brd_idx].wbf_len);
        }
        else if (inr(SSC_TNCR) == 0) {
            if (++brd_idx >= SAMPLE_BUFFERS) {
                brd_idx = 0;
            }
            outr(SSC_TNPR, (u_int) wavbuf[brd_idx].wbf_dat);
            outr(SSC_TNCR, wavbuf[brd_idx].wbf_len);
        }
    } 
}
#else
static void I2sPdcFill(void)
{
    if (brd_idx != bwr_idx) {
        if (inr(SSC_TNCR) == 0) {
            if (++brd_idx >= SAMPLE_BUFFERS) {
                brd_idx = 0;
            }
            outr(SSC_TNPR, (u_int) wavbuf[brd_idx].wbf_dat);
            outr(SSC_TNCR, wavbuf[brd_idx].wbf_len);
        }
    } 
}
#endif

/*!
 * \brief SSC interrupt handler.
 *
 * Called when PDC transmission done.
 */
static void I2sInterrupt(void *arg)
{
    I2sPdcFill();
    NutEventPostFromIrq(&i2s_que);
}

/*!
 * \brief Read value from specified AT73C213 register.
 *
 * Only a few bytes need to be transferred, thus we use polling.
 *
 * \param reg
 */
u_char At73DacReadReg(u_int reg)
{
    u_int rc;

    /* Make sure that RDRF is cleared. */
    rc = inr(DACSPI_RDR);
    /* Set MSB to 1 for read operation, followed by the 7 bit register address. */
    outr(DACSPI_TDR, 0x80FF | (reg << 8));
    /* Wait until response is available in the receive data register. */
    while ((inr(DACSPI_SR) & SPI_RDRF) == 0);
    /* Read the result from the receive data register. */
    rc = inr(DACSPI_RDR);

    return (u_char) rc;
}

/*!
 * \brief Write value to specified AT73C213 register.
 *
 * Only a few bytes need to be transferred, thus we use polling.
 *
 * \param reg
 */
void At73DacWriteReg(u_int reg, u_int val)
{
    u_int rc;

    /* Make sure that RDRF is cleared. */
    rc = inr(DACSPI_RDR);
    /* MSB is 0 for write operation, followed by the 7 bit register address. */
    outr(DACSPI_TDR, (reg << 8) | val);
    /* Wait until response is available in the receive data register. */
    while ((inr(DACSPI_SR) & SPI_RDRF) == 0);
    /* Read the result from the receive data register. Clears RDRF. */
    rc = inr(DACSPI_RDR);
}

/*!
 * \brief Configure the SPI rate.
 *
 * \param rate Baudrate. The maximum is MCK/1 and the minimum is MCK/255.
 *             If the specified rate is above the maximum or below the
 *             minimum, the maximum or minimum value resp. will be set.
 */
static int At73SpiSetRate(u_long rate)
{
    u_int divider;

    /* The SPI clock is driven by the master clock. */
    divider = (u_int) At91GetMasterClock();
    /* Calculate the SPI clock divider. Avoid rounding errors. */
    divider += (u_int) (rate / 2);
    divider /= rate;
    /* A divider value of 0 is not allowed. */
    if (divider < 1) {
        divider = 1;
    }
    /* The divider value maximum is 255. */
    else if (divider > 255) {
        divider = 255;
    }
    outr(DACSPI_CSR, (inr(DACSPI_CSR) & ~SPI_SCBR) | (divider << SPI_SCBR_LSB));

    return 0;
}

/*!
 * \brief Configure SPI master to access the AT73C213.
 */
static void At73SpiInit(void)
{
    spi_initialized = 1;
    /* Enable peripherals for SPI. */
    outr(DACSPI_ASR, DACSPI_PINS_A);
    outr(DACSPI_BSR, DACSPI_PINS_B);

    /* Disable PIO lines used for SPI. */
    outr(DACSPI_PDR, DACSPI_PINS_A | DACSPI_PINS_B);

    /* Enable SPI clock. */
    outr(PMC_PCER, _BV(DACSPI_ID));

    /* Reset SPI. */
    outr(DACSPI_CR, SPI_SWRST);

    /* Set SPI to master mode, fixed peripheral at NPCS0, fault detection disabled. */
    outr(DACSPI_MR, (90 < SPI_DLYBCS_LSB) | DACSPI_PCS | SPI_MODFDIS | SPI_MSTR);

    /* 16 bit transfer. Data changes during clock low and will be sampled on rising edges. */
    outr(DACSPI_CSR, SPI_BITS_16 | (255 << SPI_SCBR_LSB));
    /* Min. SPI clock cycle time of the AT73C213 is 100ns. */
    //At73SpiSetRate(10000);
    At73SpiSetRate(1000);

    /* Enable SPI. */
    outr(DACSPI_CR, SPI_SPIEN);
}

/*!
 * \brief Configure the SSC for I2S mode.
 *
 * \param rate Sampling rate in Hertz. 8000, 11024, 16000, 22050, 24000,
 *             32000, 44100 or 48000.
 *
 * \return 0 on success, -1 if the given rate is out of bounds.
 */
static int At73I2sSetRate(u_long rate)
{
    u_int divider;

    /* The SSC clock input equals the master clock divided by 2. */
    divider = (u_int) (At91GetMasterClock() / 2);
    /* We need to transfer 2 x 16 bit at the given sampling rate. */
    rate *= 32;
    /* Calculate the SSC clock divider. Avoid rounding errors. */
    divider += (u_int) (rate / 2);
    divider /= rate;
    /* Check limits. */
    if (divider < 1 || divider > 4095) {
        printf("Bad I2S rate\n");
        return -1;
    }
    printf("I2S Div %u\n", divider);
    outr(SSC_CMR, divider);
    return 0;
}

int At73I2sDisable(void)
{
    /* Disable all interrupts. */
    outr(SSC_IDR, 0xFFFFFFFF);

    /* Disable SSC interrupt. */
    NutIrqDisable(&sig_SSC);

    /* Disable SSC clock. */
    outr(PMC_PCDR, _BV(SSC_ID));
    /* Disable PCK0 output. */
    outr(PMC_SCDR, PMC_PCK0); //SYNC? 


    /* Reset receiver and transmitter. */
    outr(SSC_CR, SSC_SWRST | SSC_RXDIS | SSC_TXDIS);
    outr(SSC_RCMR, 0);
    outr(SSC_RFMR, 0);
    outr(SSC_PTCR, PDC_RXTDIS);
    outr(SSC_PTCR, PDC_TXTDIS);
    outr(SSC_TNCR, 0);
    outr(SSC_TCR, 0);

    return 0;
}

int At73I2sEnable(u_int rate)
{
    /* Select SSC peripheral functions. */
    outr(DACI2S_ASR, DACI2S_PINS_A);
    outr(DACI2S_BSR, DACI2S_PINS_B);

    /* Enable SSC peripheral pins. */
    outr(DACI2S_PDR, DACI2S_PINS_A | DACI2S_PINS_B);

    NutEnterCritical();
    /* Enable SSC clock. */
    outr(PMC_PCER, _BV(SSC_ID));
    /* Enable PCK0 output. */
    outr(PMC_SCER, PMC_PCK0); //SYNC?
    NutExitCritical();

    /* Set default sampling rate. */
    At73I2sSetRate(rate);

    /* Configure 16-bit stereo I2S transmit format. */
    outr(SSC_TCMR,              /* Set transmit clock mode. */
         SSC_CKS_DIV |          /* Use divided clock. */
         SSC_CKO_CONT |         /* Continous transmit clock. */
         SSC_CKG_NONE |         /* No clock gating. */
         SSC_START_FALL_RF |    /* Start transmission on falling edge at TF. */
         (1 << SSC_STTDLY_LSB) |        /* Delay start by 1 cycle. */
         ((((PCM_CHANS * PCM_BITS) / 2) - 1) << SSC_PERIOD_LSB));       /* New frame sync after 2 * (15 + 1) clocks. */
    outr(SSC_TFMR,              /* Set transmit frame mode. */
         ((PCM_BITS - 1) << SSC_DATLEN_LSB) |   /* Transmit 16 bits (1 half word). */
         SSC_MSBF |             /* Most significant bit first. */
         ((PCM_CHANS - 1) << SSC_DATNB_LSB) |   /* Transmit 2 data words per frame. */
         ((PCM_BITS - 1) << SSC_FSLEN_LSB) |    /* Length of frame sync. */
         SSC_FSOS_NEGATIVE);    /* Negative frame sync pulse. */

    /* Enable transmitter in PDC mode. */
    outr(SSC_PTCR, PDC_TXTEN);
    outr(SSC_CR, SSC_TXEN);

    return 0;
}

/*!
 * \brief Configure the SSC for I2S mode.
 */
static int At73I2sInit(void)
{
    /* Register SSC interrupt handler and enable SSC interrupt. */
    NutRegisterIrqHandler(&sig_SSC, I2sInterrupt, 0);

    At73I2sDisable();
    At73I2sEnable(SAMPLE_RATE);
    NutIrqEnable(&sig_SSC);

    return 0;
}

/*!
 * \brief Initialize AT73C213 DAC interface.
 */
int At73DacInit(void)
{
#if defined(DAC_LED_BIT)
    outr(PIOA_SODR, _BV(DAC_LED_BIT));
    outr(PIOA_PER, _BV(DAC_LED_BIT));
    outr(PIOA_OER, _BV(DAC_LED_BIT));
#endif

    /* Initialize SPI. */
    At73SpiInit();

    /* Switch all off.*/
    At73DacWriteReg(DAC_CTRL, 0);

    /* Activate codec and filter reset. */
    At73DacWriteReg(DAC_RST, 0);
    /* De-activate codec and filter reset. */
    At73DacWriteReg(DAC_RST, DAC_RST_RSTZ | DAC_RST_RESFILZ);
//#if defined (AT91SAM7X_EK)
    // 8000 16000 24000 48000
    At73DacWriteReg(DAC_CSFC, ovrsel);
//#endif
    /* Mute. */
    At73DacWriteReg(DAC_LMPG, 0x20);
    At73DacWriteReg(DAC_RMPG, 0x20);
    //At73DacWriteReg(DAC_LLOG, 0x20);
    //At73DacWriteReg(DAC_RLOG, 0x20);
    /* Precharge and master power on. */
    At73DacWriteReg(DAC_PRECH, 0xFF);
    /* Power up DAC channels. */
    At73DacWriteReg(DAC_CTRL, DAC_CTRL_ONLNOL | DAC_CTRL_ONLNOR);
    NutSleep(500);
    /* Stop precharging, keep master power on. */
    At73DacWriteReg(DAC_PRECH, DAC_PRECH_ONMSTR);
    NutSleep(1);
    /* Power up line out drivers. */
    At73DacWriteReg(DAC_CTRL, DAC_CTRL_ONDACL | DAC_CTRL_ONDACR | DAC_CTRL_ONLNOL | DAC_CTRL_ONLNOR);

    /* Set data mode to 16-bit I2S. */
    At73DacWriteReg(DAC_MISC, DAC_MISC_NBITS_16 | DAC_MISC_DINTSEL_I2S);

    /* Set default volume. */
    At73DacWriteReg(DAC_LMPG, 0x18);
    At73DacWriteReg(DAC_RMPG, 0x18);

    /* Mute line in. */
    At73DacWriteReg(DAC_LLIG, 0x11);
    At73DacWriteReg(DAC_RLIG, 0x11);

    //At73DacWriteReg(DAC_MC, 0x03);

    /* Initialize I2S. */
    At73I2sInit();

    return 0;
}

/*
 * \param rate Sampling rate in Hertz. 11025, 22050, 24000, 44100 or 48000.
 *
 */
int At73DacAdaptMasterClock(u_int rate)
{
    u_int pll_mul;
    u_int pll_div;
    u_int pck_pres;
    u_int mck_pres;
    u_int mck_div;
    u_int ovrsel;

#if defined (MCU_AT91SAM9260)

    /* PCK0 output must be set to sample rate x256 or x384. */
    switch (rate) {
    case 5513:
        /* 
         * PLLA = 18.432 MHz * (48 + 1) / 10 = 90.3168 MHz
         * PCKx = 90.3168 MHz / 32 = 2.8224 MHz (11.025 kHz * 256)
         * CPU  = 90.3168 MHz / 1 = 90.3168 MHz
         * MCK  = 90.3168 MHz / 1 = 90.3168 MHz
         * I2S  = (90.3168 MHz / 2) / 128 = 0.3528 MHz (11.025 kHz * 32)
         */
        pll_mul = 48;
        pll_div = 10;
        pck_pres = PMC_PRES_CLK_64;
        mck_pres = PMC_PRES_CLK;
        mck_div = PMC_MDIV_1;
        ovrsel = 0;
        break;
    case 8000:
        /* 
         * PLLA = 18.432MHz * (15 + 1) / 3 =  98.304 MHz
         * PCKx = 98.304 MHz / 32 = 3.072 MHz (8 kHz * 384)
         * CPU  = 98.304 MHz / 1 = 98.304 MHz
         * MCK  = 98.304 MHz / 1 = 98.304 MHz
         * I2S  = (98.304 MHz / 2) / 96 = 0.256 MHz (8 kHz * 32)
         */
        pll_mul = 15;
        pll_div = 3;
        pck_pres = PMC_PRES_CLK_32;
        mck_pres = PMC_PRES_CLK;
        mck_div = PMC_MDIV_1;
        ovrsel = DAC_CSFC_OVRSEL;
        break;
    case 11025:
        /* 
         * PLLA = 18.432 MHz * (48 + 1) / 10 = 90.3168 MHz
         * PCKx = 90.3168 MHz / 32 = 2.8224 MHz (11.025 kHz * 256)
         * CPU  = 90.3168 MHz / 1 = 90.3168 MHz
         * MCK  = 90.3168 MHz / 1 = 90.3168 MHz
         * I2S  = (90.3168 MHz / 2) / 128 = 0.3528 MHz (11.025 kHz * 32)
         */
        pll_mul = 48;
        pll_div = 10;
        pck_pres = PMC_PRES_CLK_32;
        mck_pres = PMC_PRES_CLK;
        mck_div = PMC_MDIV_1;
        ovrsel = 0;
        break;
    case 12000:
        /* 
         * PLLA = 18.432 MHz * (7 + 1) / 1 = 147.456 MHz
         * PCKx = 147.456 MHz / 32 = 4.608 MHz (12 kHz * 384)
         * CPU  = 147.456 MHz / 2 = 73.728 MHz
         * MCK  = 73.728 MHz / 1 = 73.728 MHz
         * I2S  = (73.728 MHz / 2) / 96 = 0.384 MHz (12 kHz * 32)
         */
        pll_mul = 7;
        pll_div = 1;
        pck_pres = PMC_PRES_CLK_32;
        mck_pres = PMC_PRES_CLK_2;
        mck_div = PMC_MDIV_1;
        ovrsel = DAC_CSFC_OVRSEL;
        break;
    case 16000:
        /*
         * PLLA = 18.432 MHz * (15 + 1) / 3 = 98.304 MHz
         * PCKx = 98.304 MHz / 16 = 6.144 MHz (16 kHz * 384)
         * CPU  = 98.304 MHz / 1 = 98.304 MHz
         * MCK  = 98.304 MHz / 2 = 98.304 MHz
         * I2S  = (98.304 MHz / 2) / 96 = 0.512 MHz (16 kHz * 32)
         */
        pll_mul = 15;
        pll_div = 3;
        pck_pres = PMC_PRES_CLK_16;
        mck_pres = PMC_PRES_CLK;
        mck_div = PMC_MDIV_1;
        ovrsel = DAC_CSFC_OVRSEL;
        break;
    case 22050:
        /* 
         * PLLA = 18.432 MHz * (48 + 1) / 10 = 90.3168 MHz
         * PCKx = 90.3168 MHz / 16 = 5.6448 MHz (22.05 kHz * 256)
         * CPU  = 90.3168 MHz / 1 = 90.3168 MHz
         * MCK  = 90.3168 MHz / 1 = 90.3168 MHz
         * I2S  = (90.3168 MHz / 2) / 64 = 0.705 MHz (22.05 kHz * 32)
         */
        pll_mul = 48;
        pll_div = 10;
        pck_pres = PMC_PRES_CLK_16;
        mck_pres = PMC_PRES_CLK;
        mck_div = PMC_MDIV_1;
        ovrsel = 0;
        break;
    case 24000:
        /*
         * PLLA = 18.432 MHz * (7 + 1) / 1 = 147.456 MHz
         * PCKx = 147.456 MHz / 16 = 9.216 MHz (24 kHz * 384)
         * CPU  = 147.456 MHz / 2 = 73.728 MHz
         * MCK  = 73.728 MHz / 1 = 73.728 MHz
         * I2S  = (73.728 MHz / 2) / 48 = 0.768 MHz (24 kHz * 32)
         */
        pll_mul = 7;
        pll_div = 1;
        pck_pres = PMC_PRES_CLK_16;
        mck_pres = PMC_PRES_CLK_2;
        mck_div = PMC_MDIV_1;
        ovrsel = DAC_CSFC_OVRSEL;
        break;
    case 32000:
        /*
         * PLLA = 18.432 MHz * (15 + 1) / 3 = 98.304 MHz
         * PCKx = 98.304 MHz / 8 = 12.288MHz (32 kHz * 384)
         * CPU  = 98.304 MHz / 1 = 98.304 MHz
         * MCK  = 98.304 MHz / 2 = 98.304 MHz
         * I2S  = (98.304MHz / 2) / 48 = 1.024 MHz (32 kHz * 32)
         */
        pll_mul = 15;
        pll_div = 3;
        pck_pres = PMC_PRES_CLK_8;
        mck_pres = PMC_PRES_CLK;
        mck_div = PMC_MDIV_1;
        ovrsel = DAC_CSFC_OVRSEL;
        break;
    case 44100:
        /* 
         * PLLA = 18.432 MHz * (48 + 1) / 10 = 90.3168 MHz
         * PCKx = 90.3168 MHz / 8 = 11.2896 MHz (44.1 kHz * 256)
         * CPU  = 90.3168 MHz / 1 = 90.3168 MHz
         * MCK  = 90.3168 MHz / 2 = 90.3168 MHz
         * I2S  = (90.3168MHz / 2) / 32 = 1.4112 MHz (44.1 kHz * 32)
         */
        pll_mul = 48;
        pll_div = 10;
        pck_pres = PMC_PRES_CLK_8;
        mck_pres = PMC_PRES_CLK;
        mck_div = PMC_MDIV_1;
        ovrsel = 0;
        break;
    case 48000:
        /* 
         * PLLA = 18.432 MHz * (7 + 1) / 1 = 147.456 MHz
         * PCKx = 147.456 MHz / 8 = 18.432 MHz (48 kHz * 384)
         * CPU  = 147.456 MHz / 1 = 147.456 MHz
         * MCK  = 147.456 MHz / 2 = 73.728 MHz
         * I2S  = (73.728 MHz / 2) / 24 = 1.536 MHz (48 kHz * 32)
         */
        pll_mul = 7;
        pll_div = 1;
        pck_pres = PMC_PRES_CLK_8;
        mck_pres = PMC_PRES_CLK;     
        mck_div = PMC_MDIV_2;
        ovrsel = DAC_CSFC_OVRSEL;
        break;
    default:
        return -1;
    }

    /* Disable PCK0 output. */
    outr(PMC_SCDR, PMC_PCK0);

    /* Switch to slow clock. */
    outr(PMC_MCKR, PMC_CSS_SLOW_CLK);
    while ((inr(PMC_SR) & PMC_MCKRDY) == 0);

    if (0) {
        u_int cmr = inr(SSC_CMR);

        outr(CKGR_PLLAR, _BV(29) | CKGR_OUT_0 | CKGR_PLLCOUNT | (pll_mul << CKGR_MUL_LSB) | (0 << CKGR_DIV_LSB));
        outr(SSC_CMR, 0);

        /* Wait until PLL is locked. */
        //while ((inr(PMC_SR) & PMC_LOCKA) == 0);
        outr(CKGR_PLLAR, _BV(29) | CKGR_OUT_0 | CKGR_PLLCOUNT | (pll_mul << CKGR_MUL_LSB) | (pll_div << CKGR_DIV_LSB));
        outr(SSC_CMR, cmr);

        while ((inr(PMC_SR) & PMC_LOCKA) == 0);
    }
    else {
        /* Configure PLL A. */
        outr(CKGR_PLLAR, _BV(29) | CKGR_OUT_0 | CKGR_PLLCOUNT | (pll_mul << CKGR_MUL_LSB) | (pll_div << CKGR_DIV_LSB));
        /* Wait until PLL is locked. */
        while ((inr(PMC_SR) & PMC_LOCKA) == 0);
    }
    /* Processor clock from PLLA. */
    outr(PMC_MCKR, PMC_CSS_PLLA_CLK | mck_pres | mck_div);
    while ((inr(PMC_SR) & PMC_MCKRDY) == 0);

    /* Enable PCK0 output function at pin PC1. */
    outr(PIOC_BSR, _BV(PC1_PCK0_B));
    /* Disable PC1 PIO function. */
    outr(PIOC_PDR, _BV(PC1_PCK0_B));

    if (spi_initialized) {
        u_int cmr = inr(SSC_CMR);
        u_int dly;

        NutEnterCritical();
        outr(PMC_PCKR0, PMC_CSS_PLLA_CLK | 0);
        outr(SSC_CMR, 0);
        for (dly = 0; dly < 10000; dly++) {
            _NOP();
        }
        outr(PMC_PCKR0, PMC_CSS_PLLA_CLK | pck_pres);
        outr(SSC_CMR, cmr);
        NutExitCritical();
    }
    else {
        /* Feed PCK0 from PLLA divided by pkc_pres. */
        outr(PMC_PCKR0, PMC_CSS_PLLA_CLK | pck_pres);
    }

    /* Enable PCK0 output. */
    outr(PMC_SCER, PMC_PCK0);

    if (spi_initialized) {
        At73DacWriteReg(DAC_CSFC, ovrsel);
    }
#elif defined (MCU_AT91SAM7X256)

    pck_pres = PMC_PRES_CLK_16;
    pll_mul = 538;
    pll_div = 110;

    switch (rate) {
    case 8000:
        pll_mul = 579;
        pll_div = 215;
        pck_pres = PMC_PRES_CLK_16;      /* x384 */
        break;
    case 11025:
        pck_pres = PMC_PRES_CLK_32;     /* x256 */
        break;
    case 12000:
        pll_mul = 543;
        pll_div = 136;
        pck_pres = PMC_PRES_CLK_16;      /* x384 */
        break;
    case 16000:
        pll_mul = 579;
        pll_div = 215;
        pck_pres = PMC_PRES_CLK_8;      /* x384 */
        break;
    case 22050:
        pck_pres = PMC_PRES_CLK_16;     /* x256 */
        break;
    case 24000:
        pll_mul = 543;
        pll_div = 136;
        pck_pres = PMC_PRES_CLK_8;      /* x384 */
        break;
    case 48000:
        pll_mul = 543;
        pll_div = 136;
        pck_pres = PMC_PRES_CLK_4;      /* x384 */
        break;
    case 44100:
    default:
        pck_pres = PMC_PRES_CLK_8;      /* x256 */
        break;
        //return -1;
    }
    outr(CKGR_PLLR, CKGR_USBDIV_4 | CKGR_OUT_0 | CKGR_PLLCOUNT | (pll_mul << CKGR_MUL_LSB) | (pll_div << CKGR_DIV_LSB));
    /* Wait until PLL is locked. */
    while ((inr(PMC_SR) & PMC_LOCK) == 0);

    /* Enable PIOA clock. */
    outr(PMC_PCER, _BV(PIOA_ID));
    /* Enable PCK0 output function at pin PA30. */
    outr(PIOA_BSR, _BV(PA30_PCK2_B));
    /* Disable PA30 PIO function. */
    outr(PIOA_PDR, _BV(PA30_PCK2_B));

    outr(PMC_PCKR2, PMC_CSS_PLL_CLK | pck_pres);
    /* Enable PCK2 output. */
    outr(PMC_SCER, PMC_PCK2);

    outr(PMC_MCKR, PMC_CSS_PLL_CLK);
    while ((inr(PMC_SR) & PMC_MCKRDY) == 0);

#endif

    return 0;
}

/*!
 * \brief Start transmitting audio samples.
 */
int At73DacStart(void)
{
    NutIrqDisable(&sig_SSC);
    I2sPdcFill();
    outr(SSC_IER, SSC_ENDTX);
    outr(SSC_CR, SSC_TXEN);
    NutIrqEnable(&sig_SSC);

    return 0;
}

/*!
 * \brief Wait until all buffered samples has been transmitted.
 */
int At73DacFlush(void)
{
    while (bwr_idx != brd_idx) {
        At73DacStart();
        NutEventWait(&i2s_que, 500);
    }
    return 0;
}

/*!
 * \brief Add audio samples to the AT73C213 transmit queue.
 */
int At73DacWrite(void *buf, int len)
{
    u_int idx;

#if defined(DAC_LED_BIT)
    {
        static u_int ledcnt;

        if (ledcnt++ & 2) {
            outr(PIOA_CODR, _BV(DAC_LED_BIT));
        } else {
            outr(PIOA_SODR, _BV(DAC_LED_BIT));
        }
    }
#endif

    idx = bwr_idx + 1;
    if (idx >= SAMPLE_BUFFERS) {
        idx = 0;
    }
    while (idx == brd_idx) {
        if (NutEventWait(&i2s_que, 100)) {
            At73DacStart();
        }
    }

    if (wavbuf[idx].wbf_siz < len) {
        if (wavbuf[idx].wbf_siz) {
            free(wavbuf[idx].wbf_dat);
            wavbuf[idx].wbf_siz = 0;
        }
        wavbuf[idx].wbf_dat = malloc(len * 2);
        if (wavbuf[idx].wbf_dat == NULL) {
            return -1;
        }
        wavbuf[idx].wbf_siz = len;
    }
    memcpy(wavbuf[idx].wbf_dat, buf, len * 2);
    wavbuf[idx].wbf_len = len;
    bwr_idx = idx;

    return 0;
}

int At73DacSetVolume(int left, int right)
{
    u_int lval = 0x1F;
    u_int rval = 0x1F;

    if (left > 12) {
        left = 12;
    }
    if (left < -36) {
        left = -36;
    }
    if (right > 12) {
        right = 12;
    }
    if (right < -36) {
        right = -36;
    }
    lval -= (left + 36) * 2 / 3;
    rval -= (right + 36) * 2 / 3;

    At73DacWriteReg(DAC_LMPG, lval);
    At73DacWriteReg(DAC_RMPG, rval);

    return 0;
}
